#!/usr/bin/python

#-------------------------------------------------------------------------------
#
# Copyright 2017 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------

"""
"""

#-------------------------------------------------------------------------------
#
#   Import the necessary modules
#
#-------------------------------------------------------------------------------

try:
    import argparse
    import datetime
    import errno
    import fcntl
    import logging
    import nlmanager.nlpacket
    import os
    import re
    import select
    import socket
    import struct
    import subprocess
    import threading
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")

#-------------------------------------------------------------------------------
#
#   Global Variables
#
#-------------------------------------------------------------------------------

# The handle used for logging messages
logger = logging.getLogger("cumulus-chassisd")


#-------------------------------------------------------------------------------
#
#   The chassisd_autovrf Class
#
#-------------------------------------------------------------------------------

class chassisd_autovrf:
    """
    This class communicates with other chassis cards and automatically adds them
    throughout the fabric when they are added on one node.
    """

    def __init__(self, chassis, cmdline, stopEvent):
        # Save some command line parameters which we will use later
        self.chassis        = chassis
        self.cmdargs, self.cmdline = self.ParseCmdLine(cmdline)
        self.stopEvent      = stopEvent

        # Constants
        self._VRF_INSTALLED = 1
        self._VRF_CFG_ERROR = 2
        self._IFUPDOWN2_D   = "/etc/network/interfaces.d/"

        #
        #   Netlink information structures.
        #
        #   fabByIndex a dictionary of fabric interfaces. The key is ifindex and
        #       the data is the interface name. For example:
        #       { 21 : "fp0", 22 : "fp1", ... }
        #   fabVlans a dictionary of vlan subinterface information. The key is
        #       the vlanId and the data is a dictionary. The data dictionary's
        #       keys are the vlan subinterface's ifindex and the data is an
        #       array with two elements: the vrfIfIndex (can be None is not
        #       enslaved) and the lower interface's (fabric port) ifindex.
        #        vlanId   vlanIdx   vrfIdx loIdx vlanIdx  vrfIdx loIdx
        #       { 1001  :  { 76  :  [ 110,  21 ],  77  :  [ 110,  22 ], ... },
        #         1002  :  { 96  :  [ 111,  21 ],  97  :  [ 111,  22 ], ... }, ... }
        #   vrfs a dictionary of vrf interfaces. The key is the ifindex of the
        #       vrf and the data is a two element array of the vrfId assigned to
        #       and the vrf interface name.
        #       { 110 : [ 1001, "red" ], 111 : [ 1002, "blue" ], ... }
        #
        self.netlinkLock    = threading.Lock()
        self.fabByIndex     = {}
        self.fabVlans       = {}
        self.vrfs           = {}

        # Create the threads
        self.SendingThreadObj = threading.Thread(None, self.SendingThreadT, "SendingThread")
        self.SendingThreadObj.daemon = True
        self.ReceivingThreadObj = threading.Thread(None, self.ReceivingThreadT, "ReceivingThread")
        self.ReceivingThreadObj.daemon = True

        # The only purpose of this "self pipe" is to be able to wake up the
        # receiving thread from the select.select() without having to wait for
        # a timeout.
        self.SelfPipeR, self.SelfPipeW = self.InitSelfPipe()
        if self.SelfPipeR is None or self.SelfPipeW is None:
            logger.error("AutoVRF - Could not create the self pipe.")
            self.stopEvent.set()
            return

        # Create the socket used for sending and receiving multicast packets
        self.sock, self.maddr = self.GetIpv6MulticastSock()
        if self.sock is None:
            logger.error("AutoVRF - Could not create the multicast UDP socket.")
            self.stopEvent.set()
            return

        # Start the threads
        self.SendingThreadObj.start()
        self.ReceivingThreadObj.start()


    def ParseCmdLine(self, cmdline):
        _DEFAULT_IPV6_MCAST_IP      = "ff12::5345"
        _DEFAULT_UDP_PORT           = 0x5347
        _DEFAULT_PROT_VERSION       = 1
        _DEFAULT_RECV_TIMEOUT       = 3.0
        _DEFAULT_SEND_DELAY         = 1.0
        _DEFAULT_MGMT_INTERFACE     = self.chassis.GetMgmtIf()

        parser = argparse.ArgumentParser(description="The Chassid AutoVRF Daemon")
        parser.add_argument("--vrfIpv6Mcast", default=_DEFAULT_IPV6_MCAST_IP,
                            metavar="IPV6ADDR", help="The IPv6 multicast address to use.")
        parser.add_argument("--vrfUdpPort", type=self.intBase, default=_DEFAULT_UDP_PORT,
                            metavar="UDPPORT", help="The UDP port to send/listen.")
        parser.add_argument("--vrfProtVer", type=self.intBase, default=_DEFAULT_PROT_VERSION,
                            help="The protocol version to use.")
        parser.add_argument("--vrfRecvTimeout", type=float, default=_DEFAULT_RECV_TIMEOUT,
                            metavar="SECONDS", help="Seconds to wait before receive timeout")
        parser.add_argument("--vrfSendDelay", type=float, default=_DEFAULT_SEND_DELAY,
                            metavar="SECONDS", help="Seconds between transmissions")
        parser.add_argument("--vrfMgmtIf", default=_DEFAULT_MGMT_INTERFACE,
                            metavar="INTERFACE", help="The name of the interface to use")
        args, cmdline = parser.parse_known_args(cmdline)
        return(args, cmdline)

    def intBase(self, intStr):
        '''
        Allow an integer to be specified with standard base representations:
        37 = decimal, 0x25 = hex, 045 = octal, 0b100101 = binary
        '''
        try:
            intVal = int(intStr,0)
        except:
            msg = "%s is not a valid integer" % (intStr,)
            raise argparse.ArgumentTypeError(msg)
        return intVal

    def GetIfIndex(self, ifname):
        """
        Given the name of an interface, retrieve the interface index
        """
        try:
            with open("/sys/class/net/" + ifname + "/ifindex") as f:
                return int(f.readline().strip(), 0)
        except (IOError, ValueError, TypeError):
            logger.warning("AutoVRF - Could not determine the interface index of %s" % (ifname,))
            return None

    def InitSelfPipe(self):
        """
        Create a pipe to ourselves. The only purpose of this pipe is so that it
        can be added to the read object in the select.select() call so that a
        write to this pipe, will cause the select.select() call to return. This
        is only used when exiting the program.
        """
        try:
            SelfPipeR, SelfPipeW = os.pipe()
        except os.error as e:
            logger.error("AutoVRF - Could not open self pipe (%s)" % (str(e),))
            return None, None

        try:
            flags = fcntl.fcntl(SelfPipeR, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeR, fcntl.F_SETFL, flags | os.O_NONBLOCK)
            flags = fcntl.fcntl(SelfPipeW, fcntl.F_GETFL)
            fcntl.fcntl(SelfPipeW, fcntl.F_SETFL, flags | os.O_NONBLOCK)
        except IOError as e:
            logger.error("AutoVRF - Could not get or set flags on self pipe (%s)" % (str(e),))
            os.close(SelfPipeR)
            os.close(SelfPipeW)
            return None, None
        return SelfPipeR, SelfPipeW

    def GetIpv6MulticastSock(self):
        """
        Create a multicast socket which can be used for sending and receiving
        packets.
        """
        sock = maddr = None
        mgmtAddr = self.chassis.GetMgmtAddr()
        ifidx = self.GetIfIndex(self.cmdargs.vrfMgmtIf)
        if mgmtAddr is not None and ifidx is not None:
            try:
                haddr = socket.getaddrinfo(mgmtAddr, self.cmdargs.vrfUdpPort, socket.AF_INET6, socket.SOCK_DGRAM)[0][-1]
                maddr = socket.getaddrinfo(self.cmdargs.vrfIpv6Mcast, self.cmdargs.vrfUdpPort, socket.AF_INET6, socket.SOCK_DGRAM)[0][-1]

                sock = socket.socket(socket.AF_INET6, socket.SOCK_DGRAM)

                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
                sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)

                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_LOOP, 0)
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_HOPS, 1)

                ifidx = struct.pack("I", ifidx)
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_MULTICAST_IF, ifidx)

                group = socket.inet_pton(socket.AF_INET6, maddr[0]) + ifidx
                sock.setsockopt(socket.IPPROTO_IPV6, socket.IPV6_JOIN_GROUP, group)

                sock.bind(('', self.cmdargs.vrfUdpPort))
                sock.setblocking(False)
            except socket.error as e:
                logger.error("AutoVRF - An error occurred creating the multicast socket: %s" % (str(e),))
                sock = maddr = None

        return sock, maddr

    def SignalThreads(self):
        if self.SelfPipeW is not None:
            try:
                os.write(self.SelfPipeW, 'x')
            except os.error:
                pass

    def JoinThreads(self):
        if self.SendingThreadObj is not None and self.SendingThreadObj.is_alive():
            self.SendingThreadObj.join()
        if self.ReceivingThreadObj is not None and self.ReceivingThreadObj.is_alive():
            self.ReceivingThreadObj.join()

    def NetlinkNotify(self, nlm, msg):
        """
        This function is called when Netlink Messages are received. The important
        information is saved.
        """
        if msg.family == socket.AF_UNSPEC:
            ifName = msg.get_attribute_value(msg.IFLA_IFNAME, "")
            info = msg.get_attribute_value(msg.IFLA_LINKINFO, {})
            isKnet = info.get(msg.IFLA_INFO_KIND) == "bcm_knet"
            isVlan = info.get(msg.IFLA_INFO_KIND) == "vlan"
            isVrf  = info.get(msg.IFLA_INFO_KIND) == "vrf"
            # knet interface?
            self.netlinkLock.acquire()
            if isKnet:
                # Is this a fabric port?
                if ifName.startswith("fp"):
                    # Is the interface being added?
                    present = msg.ifindex in self.fabByIndex
                    if msg.msgtype == nlmanager.nlpacket.RTM_NEWLINK and not present:
                        self.fabByIndex[msg.ifindex] = ifName
                        logger.debug("AutoVRF: Adding fabric interface { ifIdx %d : ifName %s }" % (msg.ifindex, ifName))
                    elif msg.msgtype == nlmanager.nlpacket.RTM_DELLINK and present:
                        self.fabByIndex.pop(msg.ifindex, None)
                        logger.debug("AutoVRF: Removing fabric interface { ifIdx %d : ifName %s }" % (msg.ifindex, ifName))
            # Vlan subinterface?
            elif isVlan:
                lowerIf = msg.get_attribute_value(msg.IFLA_LINK)
                isMasterVrf = info.get(msg.IFLA_INFO_SLAVE_KIND) == "vrf"
                masterIf = msg.get_attribute_value(msg.IFLA_MASTER) if isMasterVrf else None
                data = info.get(msg.IFLA_INFO_DATA, {})
                vlanId = data.get(msg.IFLA_VLAN_ID)
                # Vlan subinterface of a fabric port and has a vlan ID?
                if lowerIf in self.fabByIndex and vlanId is not None:
                    # Is the interface being added?
                    if msg.msgtype == nlmanager.nlpacket.RTM_NEWLINK:
                        vlanEntry = self.fabVlans.setdefault(vlanId, {})
                        if vlanEntry.get(msg.ifindex) != [masterIf, lowerIf]:
                            vlanEntry[msg.ifindex] = [masterIf, lowerIf]
                            logger.debug("AutoVRF: Adding vlan interface { vlanId %d : { ifIdx %d : [ vrfIdx %s,  fpIdx %s ] } }" % (vlanId, msg.ifindex, str(masterIf), str(lowerIf)))
                    elif msg.msgtype == nlmanager.nlpacket.RTM_DELLINK:
                        if vlanId in self.fabVlans:
                            self.fabVlans[vlanId].pop(msg.ifindex, None)
                            logger.debug("AutoVRF: Removing vlan interface { vlanId %d : { ifIdx %d : [ vrfIdx,  fpIdx ] } }" % (vlanId, msg.ifindex))
                            if not self.fabVlans[vlanId]:
                                self.fabVlans.pop(vlanId, None)
                                logger.debug("AutoVRF: Removing vlanId %d" % (vlanId,))
            # VRF interface?
            elif isVrf:
                data = info.get(msg.IFLA_INFO_DATA, {})
                vrfId = data.get(msg.IFLA_VRF_TABLE)
                # Is the interface being added?
                if msg.msgtype == nlmanager.nlpacket.RTM_NEWLINK:
                    if self.vrfs.get(msg.ifindex, [None]*3)[0:2] != [vrfId, ifName]:
                        if not self.isFabVrfConfigured(msg.ifindex, True):
                            if self.ConfigureFabVrf(nlm, msg.ifindex, vrfId, ifName):
                                self.vrfs[msg.ifindex] = [vrfId, ifName, self._VRF_INSTALLED]
                                logger.debug("AutoVRF: Configured new vrf interface { vrfIdx %d : [ vrfId %s, ifName %s ] }" % (msg.ifindex, str(vrfId), ifName))
                            else:
                                self.vrfs[msg.ifindex] = [vrfId, ifName, self._VRF_CFG_ERROR]
                                logger.debug("AutoVRF: Unable to configure new vrf interface { vrfIdx %d : [ vrfId %s, ifName %s ] }" % (msg.ifindex, str(vrfId), ifName))
                        else:
                            logger.debug("AutoVRF: New vrf interface already configured { vrfIdx %d : [ vrfId %s, ifName %s ] }" % (msg.ifindex, str(vrfId), ifName))
                elif msg.msgtype == nlmanager.nlpacket.RTM_DELLINK:
                    if msg.ifindex in self.vrfs:
                        vrfName = self.vrfs.get(msg.ifindex, [None]*3)[1]
                        self.UnconfigureFabVrf(nlm, vrfId, vrfName)
                        self.vrfs.pop(msg.ifindex, None)
                        logger.debug("AutoVRF: Removing vrf interface vrfIdx %d" % (msg.ifindex,))
            self.netlinkLock.release()

    def isFabVrfConfigured(self, vrfIfIndex, locked=False):
        """
        Determine if the VRF has been configured on the fabric interfaces.
        """
        return self.GetVrfVlanId(vrfIfIndex, locked)


    def ConfigureFabVrf(self, nlm, ifIndex, vrfId, vrfName):
        """
        Add the fabric configuration for a VRF. ifIndex is the interface index
        of the VRF, vrfId is the ID assigned to the VRF (which will be used as
        the VLAN ID, and vrfName is the name of the VRF. nlm is a netlink
        manager instance.
        """
        return self.ConfigureFabVrfIfs(nlm, ifIndex, vrfId, vrfName) and \
               self.ConfigureFabVrfFrr(vrfId, vrfName)

    def ConfigureFabVrfIfs(self, nlm, ifIndex, vrfId, vrfName):
        """
        Create VLAN sub-interfaces on the fabric ports with the vrfId as the
        VLAN ID. Then enslave the VLAN sub-interfaces to the VRF.
        """
        if vrfId in self.fabVlans:
            logger.debug("AutoVRF: Cannot configure VRF %s, vlan id %d in use" % (vrfName, vrfId))
            return False
        fname = "/etc/network/interfaces.d/vrf-%s-fab.intf" % (vrfName,)
        if os.path.isfile(fname):
            logger.debug("AutoVRF: Cannot configure VRF %s, file %s exists" % (vrfName, fname))
            return False
        with open(fname, "w") as f:
            f.write("#" * 80 + "\n")
            f.write("#\n")
            f.write("#   DO NOT EDIT THIS FILE!\n")
            f.write("#   File automatically created on %s by %s\n" % (datetime.datetime.now().strftime("%c"), __name__))
            f.write("#\n")
            f.write("#" * 80 + "\n\n")
            for fabIdx in range(self.chassis.GetNumFabIntfs()):
                f.write("auto fp%d.%d\n" % (fabIdx, vrfId))
                f.write("iface fp%d.%d\n" % (fabIdx, vrfId))
                f.write("    vrf %s\n\n" % (vrfName,))
        try:
            cmd = ["/sbin/ifreload", "-a"]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVRF: ifupdown2 was not configured correctly. The following command failed: %s" % (" ".join(cmd),))

        #for fabIfIndex, fabIfName in self.fabByIndex.items():
        #    vlanIfName = "%s.%d" % (fabIfName, vrfId)
        #    nlm.link_add_vlan(fabIfIndex, vlanIfName, vrfId)
        #for fabIfIndex, fabIfName in self.fabByIndex.items():
        #    vlanIfName = "%s.%d" % (fabIfName, vrfId)
        #    nlm.link_set_master(vlanIfName, ifIndex, state="up")
        return True

    def ConfigureFabVrfFrr(self, vrfId, vrfName):
        """
        Add the VRF configuration to FRR. Adds a router with the fabric VLAN
        sub-interfaces to the FRR configuration.
        """
        # Check if FRR is not running or already configured
        try:
            showrun = subprocess.check_output(["/usr/bin/vtysh", "-c", "show running-config"])
        except subprocess.CalledProcessError:
            logger.info("AutoVRF: FRR is not running. Automatic configuration will not be performed.")
            return False
        if 'router bgp 4242424242\n' not in showrun:
            logger.info("AutoVRF: FRR is not using the default fabric configuration. Automatic configuraiton will not be performed.")
            return False
        if 'router bgp 4242424242 vrf %s\n' % (vrfName,) in showrun:
            logger.info("AutoVRF: VRF %s is already configured. Automatic configuraiton will not be performed." % (vrfName,))
            return False

        # Configure the router
        cmdStr = "configure terminal\nrouter bgp 4242424242 vrf %s\n bgp router-id 42.42.42.%d\n" % (vrfName, self.chassis.GetSlotNum())
        cmdStr += " neighbor FABRIC-%d peer-group\n neighbor FABRIC-%d remote-as internal\n" % (vrfId,vrfId)
        for fabNum in range(self.chassis.GetNumFabIntfs()):
            cmdStr += " neighbor fp%d.%d interface peer-group FABRIC-%d\n" % (fabNum, vrfId, vrfId)
        try:
            cmd = ["/usr/bin/vtysh", "-c", cmdStr]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVRF: FRR was not configured correctly. The following command failed: %s" % (" ".join(cmd),))

        # Configure the IPv4 address-family
        cmdStr = "configure terminal\nrouter bgp 4242424242 vrf %s\n address-family ipv4 unicast\n" % (vrfName,)
        if self.chassis.isFabricCard():
            cmdStr += "  neighbor FABRIC-%d route-reflector-client\n" % (vrfId,)
        else:
            cmdStr += "  distance bgp 20 19 1\n"
            cmdStr += "  redistribute connected\n"
        cmdStr += " exit-address-family\n"
        try:
            cmd = ["/usr/bin/vtysh", "-c", cmdStr]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVRF: FRR was not configured correctly. The following command failed: %s" % (" ".join(cmd),))

        # Configure the IPv6 address-family
        cmdStr = "configure terminal\nrouter bgp 4242424242 vrf %s\n address-family ipv6 unicast\n" % (vrfName,)
        cmdStr += "  neighbor FABRIC-%d activate\n" % (vrfId,)
        if self.chassis.isFabricCard():
            cmdStr += "  neighbor FABRIC-%d route-reflector-client\n" % (vrfId,)
        else:
            cmdStr += "  distance bgp 20 19 1\n"
            cmdStr += "  redistribute connected\n"
        cmdStr += " exit-address-family\n"
        cmdStr += "end\nwrite file\n"
        try:
            cmd = ["/usr/bin/vtysh", "-c", cmdStr]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVrf: FRR was not configured correctly. The following command failed: %s" % (" ".join(cmd),))
        return True

    def UnconfigureFabVrf(self, nlm, vrfId, vrfName):
        """
        Remove the fabric configuration for a VRF.
        """
        self.UnconfigureFabVrfFrr(vrfName)
        self.UnconfigureFabVrfIfs(nlm, vrfId, vrfName)

    def UnconfigureFabVrfIfs(self, nlm, vrfId, vrfName):
        """
        Remove the VLAN sub-interfaces for a VRF.
        """
        fname = "/etc/network/interfaces.d/vrf-%s-fab.intf" % (vrfName,)
        if not os.path.isfile(fname):
            logger.debug("AutoVRF: Cannot configure VRF %s, file %s does not exist" % (vrfName, fname))
            return False
        os.remove(fname)
        try:
            cmd = ["/sbin/ifreload", "-a"]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVRF: ifupdown2 was not configured correctly. The following command failed: %s" % (" ".join(cmd),))
        #vlanIfs = self.fabVlans.get(vrfId, {})
        #for vlanIfIndex in vlanIfs:
        #    nlm.link_del(vlanIfIndex)

    def UnconfigureFabVrfFrr(self, vrfName):
        """
        Remove the FRR configuration for a VRF.
        """
        cmdStr = "configure terminal\nno router bgp 4242424242 vrf %s\nend\nwrite file\n" % (vrfName,)
        try:
            cmd = ["/usr/bin/vtysh", "-c", cmdStr]
            subprocess.check_call(cmd)
        except subprocess.CalledProcessError:
            logger.error("AutoVrf: FRR was not configured correctly. The following command failed: %s" % (" ".join(cmd),))

    def GetVrfVlanId(self, vrfIfIndex, locked=False):
        """
        Determine the VLAN ID being used for a VRF on the fabric interfaces.
        All fabric ports must have a vlan subinterface and all of them must
        be enslaved to the VRF.
        """
        retVal = None
        if not locked:
            self.netlinkLock.acquire()
        numFabs = len(self.fabByIndex)
        if numFabs != 0 and vrfIfIndex in self.vrfs:
            for vlanId, subIfs in self.fabVlans.iteritems():
                if len(subIfs) != numFabs:
                    continue
                diffFound = False
                for idxes in subIfs.itervalues():
                    if idexs[0] != vrfIfIndex:
                        diffFound = True
                        break
                if not diffFound:
                    retVal = vlanId
                    break
        if not locked:
            self.netlinkLock.release()
        return retVal

    def SendingThreadT(self):
        """
        This function is just a wrapper around the thread which catches any
        unhandled exceptions and prints them to the log file. Otherwise these
        exceptions would not be logged anywhere.
        """
        try:
            self.SendingThread()
        except Exception as e:
            logger.exception(e)
            self.stopEvent.set()

    def SendingThread(self):
        """
        This function sends out a multicast packet with the current information if
        we are the sender.
        """
        logger.debug("AutoVRF - Beginning execution of the thread SendingThread")
        while not self.stopEvent.wait(self.cmdargs.vrfSendDelay):

            pass

        logger.debug("AutoVRF - Finished execution of the thread SendingThread")

    def ReceivingThreadT(self):
        """
        This function is just a wrapper around the thread which catches any
        unhandled exceptions and prints them to the log file. Otherwise these
        exceptions would not be logged anywhere.
        """
        try:
            self.ReceivingThread()
        except Exception as e:
            logger.exception(e)
            self.stopEvent.set()

    def ReceivingThread(self):
        """
        This function handles the reception of a packet from a neighbor.
        """
        logger.debug("AutoVRF - Beginning execution of the thread ReceivingThread")
        while not self.stopEvent.is_set():

            try:
                # Wait until a packet comes in or a timeout
                ready = select.select([self.sock, self.SelfPipeR], [], [])
            except select.error as e:
                code, msg = e.args
                if code == errno.EINTR:
                    continue
                else:
                    logger.error("AutoVRF - An error occurred during the select system call: %s" % (str(e),))
                    continue

            # Were we woken up just to exit?
            if self.SelfPipeR in ready[0]:
                logger.debug("AutoVRF - Received data on self pipe. Preparing to exit thread.")
                while True:
                    try:
                        os.read(self.SelfPipeR, 1)
                    except OSError as e:
                        # Is the pipe totally drained? Exit.
                        if e.errno == errno.EAGAIN:
                            break
                        else:
                            logger.error("AutoVRF - An error occured while draining the self pipe: %s" % (str(e),))
                            break
            elif self.sock in ready[0]:
                try:
                    recvData = self.sock.recvfrom(1024)
                except socket.error as e:
                    logger.warning("AutoVRF - Unable to receive data from the multicast socket: %s" % (str(e),))
                    break

                logger.debug("AutoVRF - Received a packet.")

        logger.debug("AutoVRF - Finished execution of the thread ReceivingThread")


if __name__ == '__main__':
    pass
