#! /usr/bin/python
#-------------------------------------------------------------------------------
#
# Copyright 2017 Cumulus Networks, inc  all rights reserved
#
#-------------------------------------------------------------------------------
#
"""
This script provides a Netlink interface for the cumulus-chassisd service. Is
is basically an interface to the python nlmanager.
"""

#-------------------------------------------------------------------------------
#
#   Import the necessary modules
#
#-------------------------------------------------------------------------------

try:
    import argparse
    import logging
    from   nlmanager.nllistener import NetlinkManagerWithListener
    import nlmanager.nlpacket
    import Queue
    import threading
except ImportError, e:
    raise ImportError (str(e) + "- required module not found")


#-------------------------------------------------------------------------------
#
#   Global Variables
#
#-------------------------------------------------------------------------------

# The handle used for logging messages
logger = logging.getLogger("cumulus-chassisd")


#-------------------------------------------------------------------------------
#
#   The Chassis Netlink Class
#
#-------------------------------------------------------------------------------

class chassisd_netlink:

    # Event code constants
    _GET_ALL_LINKS          = 1
    _SHUTDOWN               = 2
    _SERVICE_NETLINK_QUEUE  = 'SERVICE_NETLINK_QUEUE'

    def __init__(self, chassis, cmdline, stopEvent, distrib):
        self.chassis = chassis
        self.stopEvent = stopEvent
        self.distrib = distrib
        self.cmdargs, self.cmdline = self.ParseCmdLine(cmdline)

        self.nll = None
        self.netlinkThread = threading.Thread(None, self.NetlinkThreadT, "NetlinkThreadT")
        self.netlinkThread.daemon = True
        self.netlinkThread.start()

    def __del__(self):
        # Signal the update thread to wake up and exit
        self.SignalThreads()
        self.JoinThreads()

    def ParseCmdLine(self, cmdline):
        parser = argparse.ArgumentParser(description="The Netlink Manager Interface")
        args, cmdline = parser.parse_known_args(cmdline)
        return(args, cmdline)

    def SignalThreads(self):
        if self.nll:
            self.nll.workq.put((chassisd_netlink._SHUTDOWN, None))

    def JoinThreads(self):
        if self.nll:
            self.nll.listener.join()
            self.nll = None
        self.netlinkThread.join()

    def NetlinkThreadT(self):
        """
        This function is just a wrapper around the thread which catches any
        unhandled exceptions and prints them to the log file. Otherwise these
        exceptions would not be logged anywhere.
        """
        try:
            self.NetlinkThread()
        except Exception as e:
            logger.exception(e)
            self.stopEvent.set()

    def NetlinkThread(self):
        """
        This function operates as an independent thread of the script.
        """

        logger.debug("Beginning execution of the thread NetlinkThread")

        # These are the groups we are interesting in listening to.
        groups = nlmanager.nlpacket.RTMGRP_LINK

        # Instantiate a netlink manager with a listener.
        self.nll = ChassisNlmWListener(groups, logger, self.distrib)

        # Enable any message debugging here.
        #self.nll.debug_neighbor(True)
        #self.nll.debug_link(True)
        nlmanager.nlpacket.log = logger

        # Prime the pump by getting all interfaces
        self.nll.workq.put((chassisd_netlink._GET_ALL_LINKS, None))

        # This is the main thread loop.
        while not self.stopEvent.is_set():

            try:
                (event, options) = self.nll.workq.get(True, 60*60*24) # If timeout is None, signals are blocked
            except Queue.Empty:
                continue

            if event == chassisd_netlink._GET_ALL_LINKS:
                self.nll.get_all_links()
            elif event == chassisd_netlink._SERVICE_NETLINK_QUEUE:
                self.nll.service_netlinkq()
            elif event == chassisd_netlink._SHUTDOWN:
                logger.debug("NetlinkThread: shutting things down")
            else:
                raise Exception("Unsupported workq event %s" % event)

        self.nll.listener.shutdown_event.set()
        logger.debug("Finished execution of the thread NetlinkThread")


class ChassisNlmWListener(NetlinkManagerWithListener):
    """
    The main job of this class is to override the functions of the
    NetlinkManagerWithListener class to do the things needed when items are
    placed on the netlink queue.
    """
    def __init__(self, groups, logger, distrib):
        super(ChassisNlmWListener, self).__init__(groups, use_color=False)
        self.distrib = distrib

    def rx_rtm_newlink(self, msg):
        self.distrib(self, msg)

    def rx_rtm_dellink(self, msg):
        self.distrib(self, msg)

